# -*- coding: utf-8 -*-
#Function:  1) Parses TM1 tules and process files
#           2) Creates a networkx graph of model objects
#           3) Outputs this graph in simple csv file
#Author: Yuri Kudryavcev email mail@ykud.com
# revision history:
# 12/07/2013: changed dimension_subset definition, excluded amy_TM1_name_without_commas from the list of possible dimension names
#			  changed to suit multiple FORMAT=100 in the beginning of the rule file

import codecs
import re
import os
import sys
import re
import string
import getopt
import time
import datetime
import networkx as nx
import pydot
from pyparsing import Word, lineEnd, lineStart, line, lineno, col, Regex,Literal,CaselessLiteral, Optional,oneOf,OneOrMore, nums, ZeroOrMore,\
     Group,delimitedList,Suppress, Forward, quotedString,alphas,SkipTo,countedArray, stringEnd,ParserElement,ParseResults, Empty,NotAny,QuotedString,sglQuotedString,\
     ParseException

# Global variable for code parsing
global current_node_name
global reference_list
global feeders_part_of_rule
global parsing_pro_file #need this flag for attrs drawing
global TI_variable_list # list to track some basic variable assigment
#
# Constants
TM1_RULE_FILE_FUNCTIONS = '''DB ATTRN ATTRS ISLEAF UNDEF UNDEFVALS DATE DATES DAY DAYNO MONTH NOW TIME TIMST TIMVL TODAY YEAR ConsolidateChildren ConsolidatedMin
ConsolidatedMax ConsolidatedAvg ConsolidatedCount ConsolidatedCountUnique DIMNM DIMSIZ DNEXT DNLEV TABDIM DIMIX DTYPE ELCOMP ELCOMPN ELISANC ELISCOMP ELISPAR ELLEV
ELPAR ELPARN ELWEIGHT ELISPAR FV PAYMT PV ABS ACOS ASIN ATAN COS EXP INT ISUND LN LOG MAX MIN MOD RAND ROUND ROUNDP SIGN SIN SQRT TAN CAPIT CHAR CODE DELET FILL INSRT
LONG LOWER NUMBR SCAN STR SUBST TRIM UPPER'''
TM1_TI_ASCII_FUNCTIONS = 'ASCIIDelete SetInputCharacterSet SetOutputCharacterSet TextOutput ASCIIOutput'
TM1_TI_ATTRIBUTE_MANIPULATION_FUNCTIONS = 'AttrDelete AttrInsert AttrPutN AttrPutS SwapAliasWithPrincipalName'
TM1_TI_CHORE_FUNCTIONS = 'ChoreQuit SetChoreVerboseMessages'
TM1_TI_DIMENSION_FUNCTIONS = '''DimensionCreate DimensionDeleteAllElements DimensionDestroy DimensionEditingAliasSet DimensionElementComponentAdd DimensionElementComponentDelete DimensionElementDelete
DimensionElementInsert DimensionElementPrincipalName DimensionElementInsertByAlias DimensionExists DimensionSortOrder'''
TM1_TI_CUBE_FUNCTIONS = ''' AddCubeDependency CellPutProportionalSpread CellPutS CellPutN CellGetN CellGetS CellIsUpdateable CubeClearData
CubeCreate CubeDestroy CubeExists CubeGetLogChanges CubeSetConnParams CubeSetLogChanges CubeUnload'''
TM1_TI_ODBC_FUNCTIONS = 'ODBCClose ODBCOpen ODBCOPENEx ODBCOutput SetODBCUnicodeInterface'
TM1_TI_PROCESS_FUNCTIONS = 'ExecuteProcess ExecuteCommand GetProcessErrorFileDirectory GetProcessErrorFilename GetProcessName ItemReject ProcessExitNormal'
TM1_TI_RULES_FUNCTIONS = 'CubeProcessFeeders DeleteAllPersistentFeeders RuleLoadFromFile'
TM1_TI_SANDBOX_FUNCTIONS = 'ServerActiveSandboxGet ServerActiveSandboxSet GetUseActiveSandboxProperty SetUseActiveSandboxProperty'
TM1_TI_SECURITY = '''AddClient AddGroup AssignClientToGroup AssignClientPassword DeleteClient DeleteGroup ElementSecurityGet
ElementSecurityPut RemoveClientFromGroup'''
TM1_TI_SERVER_FUNCTIONS = 'BatchUpdateFinish BatchUpdateFinishWait DisableBulkLoadMode EnableBulkLoadMode ServerShutdown'
TM1_TI_SUBSET_FUNCTIONS = '''SubsetAliasSet SubsetCreate SubsetCreateByMDX SubsetDeleteAllElements SubsetDestroy SubsetElementDelete SubsetElementInsert
SubsetExists SubsetExpandAboveSet SubsetFormatStyleSet SubsetGetElementName SubsetGetSize SubsetIsAllSet'''
TM1_TI_VIEW_FUNCTIONS = '''PublishView ViewColumnDimensionSet ViewColumnSuppressZeroesSet ViewConstruct ViewCreate ViewDestroy ViewExists
ViewExtractSkipCalcsSet ViewExtractSkipRuleValuesSet ViewExtractSkipZeroesSet ViewRowDimensionSet ViewRowSuppressZeroesSet
ViewSubsetAssign ViewSuppressZeroesSet ViewTitleDimensionSet ViewTitleElementSet ViewZeroOut
ViewSetSkipZeroes ViewSetSkipCalcs ViewSetSkipRuleValues'''
TM1_TI_MISC_FUNCTION = '''AddInfoCubeRestriction DataSourceSAPUsingRoleAuths DataSourceSAPUsingTexts Expand FileExists NumberToString NumberToStringEx RefreshMdxHierarchy StringToNumber StringToNumberEx TM1User WildcardFileSearch'''
# Some TI functions may be called without () -- parameter specification. It's important to distinguish them from others since if all parameters list are made optional
# variable names (like "day") can be confused for function names
TM1_TI_KEYWORDS = 'ItemSkip ProcessBreak ProcessError SecurityRefresh ProcessQuit SaveDataAll BatchUpdateStart'
# Process drawing constants
TI_PRO_COLOR = 'blue'
TI_PRO_ARROW_STYLE = 'solid'
#
class Reference(dict):
    '''Store all types of references to other cubes/dimensions found in rules file
        References are then translated into edges of resulting graph
        Required attributes: source_node, target_node, id'''
    def __init__(self, st, locn, tokString, attributes = None):
        self["location"] = locn
        #self["source_line"] = line(locn,st)
        self["line_number"] = lineno(locn,st)
        self["column_number"] = col(locn,st)
        # adding other attributes
        if attributes:
            self.update(attributes)
    def __str__(self):
        return str(self.items())
#
def unique_node_name(string, node_type='cub'):
    ''' Return unique string representation.
    TM1 ignores spaces and case, so it's convenient to do the same for cubes\dimensions names.'''
    #print 'node name %s , node type %s'%(string,node_type)
    return ('"%s_%s"'%(node_type,string.replace('.','').replace(' ','').replace('\n','').replace('\\','').lower())).encode('utf-8')
#
def concatenate_array(array):
	'''Return concatenated array contents as a single string'''
	str = ''
	for part in array:
		if isinstance(part, basestring):
			 str = str + part
		else: 
			str = str + (concatenate_array(part))
	return str
#
def set_feeders_flag_on(s,l,t):
    """Setting feeders file part of file to true
    """
    global feeders_part_of_rule
    feeders_part_of_rule = True  
#
def add_Rules_function(s,l,t):
    '''Selecting functions we'll add edges for and adding needed reference attributes'''
    global current_node_name
    global reference_list
    global feeders_part_of_rule
    global parsing_pro_file
    if t[0].lower() == 'db':
        #if we've got strings on input, not nested functions
        if not isinstance(t[1],ParseResults):
            if feeders_part_of_rule:
                # if it's a feeder, we change source <->target and line style
                attributes = {"id":'"feeder_%s--->%s"'%(unique_node_name(current_node_name,'cub').replace('"',''),unique_node_name(t[1],'cub').replace('"','')),
                              "source_node_name":current_node_name,"source_node_id":unique_node_name(current_node_name,'cub'),
                              "target_node_name":t[1],"target_node_id":unique_node_name(t[1],'cub'),
                              "arrow_style":"dashed","type":"DB","feeder":feeders_part_of_rule,"color":"black"}
            else:
                attributes = {"id":'"db_%s--->%s"'%(unique_node_name(t[1],'cub').replace('"',''),unique_node_name(current_node_name,'cub').replace('"','')),
                              "source_node_name":t[1],"source_node_id":unique_node_name(t[1],'cub'),
                              "target_node_name":current_node_name,"target_node_id":unique_node_name(current_node_name,'cub'),
                              "arrow_style":"solid","type":"DB","feeder":feeders_part_of_rule,"color":"black"}
            reference_list.append(Reference(s,l,t[1],attributes))
    elif t[0].lower() in ['attrs','attrn']:
        if not isinstance(t[1],ParseResults) and not isinstance(t[-1],ParseResults):
            if parsing_pro_file:
                attributes =  {"id":'"attr_%s--->%s"'%(unique_node_name(t[1],'dim').replace('"',''),unique_node_name(current_node_name,'cub').replace('"','')),
                           "source_node_name":t[1],"source_node_type":'dim',
                           "target_node_name":current_node_name,"target_node_type":'pro',
                           "arrow_style":"dotted", "dimension_attribute":t[-1],"type":"ATTR for Attribute %s"%(t[-1]),"color":TI_PRO_COLOR}
            else:
                attributes =  {"id":'"attr_%s--->%s"'%(unique_node_name(t[1],'dim').replace('"',''),unique_node_name(current_node_name,'cub').replace('"','')),
                           "source_node_name":t[1],"source_node_id":unique_node_name(t[1],'dim'),
                           "target_node_name":current_node_name,"target_node_id":unique_node_name(current_node_name,'cub'),
                           "arrow_style":"dotted", "dimension_attribute":t[-1],"type":"ATTR","color":"black"}
            reference_list.append(Reference(s,l,str(t),attributes))
#
def add_TI_function(s,l,t):
    '''Selecting TI functions to draw and adding Reference attributes'''
    global reference_list
    global current_node_name
    # Source and target and attribute names can be variables that will be substituted for real values while processing file as whole
    # Adding source and target types for correct node creation
    # Dimension manipulation
    #print t[0] #, t[1]
    
    if  t[0].lower()  in TM1_TI_DIMENSION_FUNCTIONS.lower().split(' '):
    	#print t[0], t[1]
        if len(t[1]) == 1:
            attributes =  {"source_node_name":current_node_name, "source_node_type":"pro"
                           ,"target_node_name":t[1][0],"target_node_type":'dim',
                           "arrow_style":TI_PRO_ARROW_STYLE, "type":"%s"%(t[0]),"color":TI_PRO_COLOR}      
    # Subset Manipulation
    if  (t[0].lower()  in 'SubsetAliasSet SubsetCreate SubsetDeleteAllElements SubsetDestroy SubsetElementDelete SubsetElementInsert'.lower().split(' ')):
        if len(t[1]) == 1:
            attributes =  {"source_node_name":current_node_name, "source_node_type":"pro"
                           ,"target_node_name":t[1][0],"target_node_type":'dim',"subset_name":t[2][0],
                           "arrow_style":TI_PRO_ARROW_STYLE, "type":"%s for subset %s"%(t[0],t[2][0]),"color":TI_PRO_COLOR} 
    # ViewZeroOut
    elif t[0].lower() == 'ViewZeroOut'.lower():
        if len(t[1]) == 1:
            attributes =  {"source_node_name":current_node_name, "source_node_type":"pro"
                           ,"target_node_type":'cub',"target_node_name":t[1][0].rstrip(' '),
                           "arrow_style":TI_PRO_ARROW_STYLE, "type":t[0],"color":TI_PRO_COLOR}
    # attribute functions
    elif t[0].lower() == 'attrputs' or t[0].lower() == 'attrputn':
        #print t[2]
        if len(t[2]) == 1:
            attributes =  {"source_node_name":current_node_name, "source_node_type":"pro",
                           "target_node_name":t[2][0],"target_node_type":'dim',
                       "arrow_style":"solid", "type":"AttrPut attribute %s"%(t[-1][0]),"color":TI_PRO_COLOR,"dimension_attribute":t[-1][0]}
    elif t[0].lower() == 'attrdelete':
        #print t[1]
        if (len(t[1]) == 1):
            attributes =  {"source_node_name":current_node_name, "source_node_type":"pro",
                        "target_node_name":t[1][0],"target_node_type":'dim',
                       "arrow_style":"solid", "type":"AttrDelete attribute %s"%(t[-1][0]),"color":TI_PRO_COLOR,"dimension_attribute":t[-1][0]}
    elif t[0].lower() == 'attrinsert':
         if len(t[1]) == 1:
            attributes =  {"source_node_name":current_node_name, "source_node_type":"pro","target_node_name":t[1][0],"target_node_type":'dim',
                       "arrow_style":"solid", "type":"AttrInsert attribute %s"%(t[-2][0]),"color":TI_PRO_COLOR,"dimension_attribute":t[-2][0]}
    elif t[0].lower() == 'SwapAliasWithPrincipalName'.lower():
        if  len(t[1]) == 1:
             attributes =  {"source_node_name":current_node_name, "source_node_type":"pro",
                            "target_node_name":t[1][0],"target_node_type":'dim',
                       "arrow_style":"solid", "type":"SwapAliasWithPrincipalName Alias %s"%(t[-2][0]),"color":TI_PRO_COLOR}
    # TextOutput
    elif t[0].lower() in 'TextOutPut AsciiOutPut'.lower().split(' '):
        
        if len(t[1]) == 1 & isinstance(t[1][0],basestring):
            #print t[1][0]
             
            attributes =  {"source_node_name":current_node_name, "source_node_type":"pro",
                           "target_node_name":t[1][0],"target_node_type":'txt',
                           "arrow_style":"solid", "type":"TextOut","color":TI_PRO_COLOR}
        else: # some sort of calculated file name
            
            #print t[1],t[1][0]
            attributes =  {"source_node_name":current_node_name, "source_node_type":"pro",
                            "target_node_name":'Calculated filename %s'% (concatenate_array(t[1][0])),"target_node_type":'txt',
                            "arrow_style":"solid", "type":"TextOut","color":TI_PRO_COLOR}
                                                        
    # Execute process
    elif t[0].lower() == 'executeprocess':
        if len(t[1]) == 1:
            attributes =  {"source_node_name":current_node_name, "source_node_type":"pro",
                           "target_node_name":t[1][0],"target_node_type":'pro',
                       "arrow_style":"solid", "type":"ExecuteProcess","color":TI_PRO_COLOR}
    # CellPut
    elif t[0].lower() in 'CellPutS CellPutN CellPutProportionalSpread'.lower().split(' '):
        if len(t[2]) == 1:
            attributes =  {"source_node_name":current_node_name, "source_node_type":"pro",
                        "target_node_name":t[2][0],"target_node_type":'cub',
                       "arrow_style":TI_PRO_ARROW_STYLE, "type":"CellPut","color":TI_PRO_COLOR}
    elif t[0].lower() in 'CellGets CellGetN'.lower().split(' '):
        if len(t[1]) == 1:
            attributes =  {"source_node_name":t[1][0], "source_node_type":'cub',
                        "target_node_name":current_node_name,"target_node_type":"pro",
                       "arrow_style":TI_PRO_ARROW_STYLE, "type":"CellGet","color":TI_PRO_COLOR}
    else:
        return
    reference_list.append(Reference(s,l,str(t),attributes))
    
#
def add_TI_variable_to_list(s,l,t):
    '''Recording string variables to a list, to do some basic substition in formulas later'''
    global TI_variable_list
    # recording only simple strings and if they are no empty
    if len(t)==2:
        if isinstance(t[1], ParseResults):
            TI_variable_list.update({t[0].rstrip(' ').lstrip(' '):"Calculation"})
        else:
            TI_variable_list.update({t[0].rstrip(' ').rstrip(' '):t[1]})
#
def add_TI_param_to_list(s,l,t):
    '''Recording param names to a list, to do some basic substition in formulas later, if required'''
    global TI_variable_list
    TI_variable_list.update({t[0]:"PARAMETER"})    
#
def caseless_literal_with_line_ends(string):
    '''Create a Parse expression for string, allowing linebreaks anywhere in the string
    The only way to work with function names split into multiple strings
    Given a space separated string - returns an |(or)ed parse expression list'''
    if len(string.split(' ')) > 1:
        parse_expr = caseless_literal_with_line_ends(string.split(' ')[0])
        for x in string.split(' ')[1:]:
            if len(x)>0:
                if caseless_literal_with_line_ends(x):
                    parse_expr = caseless_literal_with_line_ends(x)|parse_expr
        if parse_expr:
            return parse_expr
    else:
        if len(string.lstrip(' ')) > 0:
            expression = ""
            for char in string:
                expression += "(%s|%s)(\\n|\\r\\n)*" % (char.lower(),char.upper())
            # removing last (\\n|\\r\\n)* not to let it eat out line breaks
            return Regex(expression[:-10])
        else: return None
#
class TM1Syntax():
    '''
    Defining TM1 rules and process files syntax. Since rules statement are can be included in pro files, making it a single definition
    '''
    rule_file_syntax = None
    pro_file_syntax = None
    test = None
    def __init__(self):
        # RULE FILE SYNTAX
        # alphabet contains all unicode chars without punctuation
        #english_chars = u''.join(unichr(x) for x in xrange(0x041, 0x05B)) +  u''.join(unichr(x) for x in xrange(0x061, 0x07B))
        #russian_chars = u''.join(unichr(x) for x in xrange(0x410, 0x450)) + unichr(0x451) 
        #extended_chars = english_chars + russian_chars
        extended_chars = u''.join(unichr(c) for c in xrange(0, 65536)
                                  if (unichr(c) not in '" &@%|?.,:;=[]<>^_!()+-*\/\n\r') and unichr(c) <> "'")
        # Some set of punctuation constants
        LBRACK,RBRACK,LPAREN,RPAREN,LBRACE,RBRACE,COMMA,QUOTE,DQUOTE,COLON,SEMICOLON,EQUALS = map(Suppress,"[](){},'\":;=")
        # comment is a line starting with any number of whitespace or tabs, followed by # sign. All from # sign is ignored
        comment = Regex(r"(^|( *(\n|\r\n))+) *#.*").leaveWhitespace().setResultsName('comment')
        #comment.setDebug()
        # all the ways a variable or cube\dimension can be caled
        # See chapter 1 of TM1 Developer Guide for naming conventions
        # http://www-01.ibm.com/support/docview.wss?uid=swg27024713
        # need to define PRO variable separately, cause it may contain % sign, but Rules cannot, since it's a logical OR operator
        # It seems they can be violated, you can use '/','%','*' in element\cube names, for example
        multi_line_any_TM1_name_without_comma = Forward()
        # a dimension reference can be writen after ! sign without quotes if it doesn't contain any reserved characters.        multi_line_any_TM1_name_without_comma = Forward()
        any_TM1_name_without_comma =  Word( extended_chars + nums + '#}. -_$[]').setName('TM1_Name')
        split_any_TM1_name_without_comma = any_TM1_name_without_comma + Suppress(lineEnd)
        multi_line_any_TM1_name_without_comma << ((split_any_TM1_name_without_comma + multi_line_any_TM1_name_without_comma) | any_TM1_name_without_comma).setName('TM1_Name_wo_comma_and_lines')
        multi_line_any_TM1_name_without_comma.addParseAction(lambda t: ''.join(t))
        # Dimensions and cubes are quoted when referenced
        single_quoted_string = QuotedString("'",None,"''",True,True)
        #single_quoted_string = "'" + multi_line_any_TM1_name_without_comma  + "''"
        single_quoted_string.addParseAction(lambda t: ''.join(t[0].split()))
        # Area definition
        # Area is a delimited list of dimension references, subsets and elements. It may be empty, or prefixed with dimension name with colon
        dimension_subset = Optional (single_quoted_string + COLON) + (LBRACE + Optional(delimitedList(single_quoted_string))  + RBRACE).setResultsName("subset")
        #dimension_subset_with_dim_reference = single_quoted_string + COLON + (LBRACE + Optional(delimitedList(single_quoted_string))  + RBRACE).setResultsName("subset")
        # Dimension element
        # element and dimension reference may not be quoted, although documentation says otherwise 
        dimension_element_specification = single_quoted_string|multi_line_any_TM1_name_without_comma
        dimension_element = (dimension_element_specification + COLON + dimension_element_specification).setResultsName("dimension element")
        area = (LBRACK + Optional(delimitedList(dimension_subset|dimension_element|single_quoted_string))  + RBRACK).setResultsName("area")
        target_area = (area).setResultsName("target_area")
        # Formula components
        # Suppressing line breaks in numbers is tricky, because if i just use Word(nums + '\r\n') all the following line-breaks will be swallowed and
        # comments parsing will fail to recognize new line.
        numeric_constant = (Word(nums + '.') + ZeroOrMore(Suppress(lineEnd) + Word(nums + '.'))).setResultsName("num_const")
        numeric_constant.setParseAction(lambda t: ''.join(t))
        # Specific function definitions for the ones we'll be drawing links for
        if_function = Forward()
        all_Rules_functions = Forward()
        arithmetic_expression = Forward()
        # grouping all different functions in one place, since arithmetic expressions have end-point types(variables, numerics) and process formulas without
        # assigment should not
        arithm_term = ((LPAREN + arithmetic_expression + RPAREN) |
                        Group(if_function)|Group(all_Rules_functions)|
                        ('!' + multi_line_any_TM1_name_without_comma)|
                        ('!' + single_quoted_string)| 
                        caseless_literal_with_line_ends("STET")|caseless_literal_with_line_ends("CONTINUE")|
                        area|
                        single_quoted_string | numeric_constant | multi_line_any_TM1_name_without_comma
                      )
        # Not using pypasing.operatorPrecedence since it's uninmportant for drawing links and simple works faster
        arithmetic_expression<< (Optional('-') +  arithm_term 
                                + Optional(Group(OneOrMore(oneOf('* \\ / * ^ | + -')).setResultsName('arithm_operator')  + Group(arithmetic_expression)))
                                )
        # if function
        logical_expression = Forward()
        # Once again (as with numeric constants) let's assume only one linebreak per operator definition
        logical_operator = Optional('@') + Word('><=') + Optional(Suppress(lineEnd) + Word('><='))
        logical_operator.setParseAction(lambda t: ''.join(t))
        logical_term = (LPAREN + logical_expression + RPAREN)|(arithmetic_expression + logical_operator + arithmetic_expression) 
        logical_expression << (Optional('~') + logical_term  \
                                + Optional(oneOf('& %').setResultsName('logical_operator') + Group(logical_expression)))
                                                     
        if_argument = (arithmetic_expression)
        if_function  << (Suppress(caseless_literal_with_line_ends("IF")) +  LPAREN + logical_expression + COMMA  + if_argument.setResultsName('if_true') + ','\
                        + if_argument.setResultsName('if_false') +   RPAREN)

        all_Rules_functions << (caseless_literal_with_line_ends(TM1_RULE_FILE_FUNCTIONS.replace('\n',' ').replace('  ',''))
                            #oneOf(TM1_RULE_FILE_FUNCTIONS.replace('\n',' ').replace('  ',''),caseless=True)
                             + LPAREN + Optional(delimitedList(arithmetic_expression)) + RPAREN ).setParseAction(add_Rules_function)
        #
        # Assigment (=C N and etc)
        assign =(Optional(EQUALS) + oneOf('C S N',caseless=True) + COLON)|Literal ('=')|COLON
        formatDefinition = caseless_literal_with_line_ends('FORMAT==') + Word (nums) + lineEnd
        SKIPCHECK = caseless_literal_with_line_ends('SKIPCHECK') + SEMICOLON
        tm1formula = Optional(target_area) + assign + arithmetic_expression + SEMICOLON + Optional(Suppress(lineEnd))
        feedstrings_header = caseless_literal_with_line_ends('FEEDSTRINGS') + Optional(SEMICOLON)
        feeders_header = caseless_literal_with_line_ends('FEEDERS') + Optional(SEMICOLON)
        feeders_header.setParseAction(set_feeders_flag_on)
        feeders_formula = Optional(target_area) + '=>' + delimitedList(arithmetic_expression) + SEMICOLON + Optional(Suppress(lineEnd))
        # 12/07/2013: changed to match multiple FORMAT= lines in rules header 
        rule_file = ZeroOrMore(formatDefinition) + Optional (SKIPCHECK)  + ZeroOrMore(tm1formula) + Optional(feeders_header + ZeroOrMore(feeders_formula))

        rule_file.ignore(feedstrings_header)
        rule_file.ignore(comment)
        #rule_file.enablePackrat() #-- actually slows things down
        self.rule_file_syntax =  rule_file

        # PROCESS FILE SYNTAX
        # Repeating definitions for rules functions, because they can mix TI and Rule function calls and keeping them separate speeds up rules file processing
        # Variables can have % sign in their names in Process files, but not in Rule files
        multi_line_any_TM1_TI_name_without_comma = Forward()
        # a dimension reference can be writen after ! sign without quotes if it doesn't contain any reserved characters.        multi_line_any_TM1_name_without_comma = Forward()
        any_TM1_TI_name_without_comma =  Word( extended_chars + nums + '#}. -_$[]%').setName('TM1_TI_Name')
        split_any_TM1_TI_name_without_comma = any_TM1_TI_name_without_comma + Suppress(lineEnd)
        multi_line_any_TM1_TI_name_without_comma << ((split_any_TM1_TI_name_without_comma + multi_line_any_TM1_TI_name_without_comma) | any_TM1_TI_name_without_comma).setName('TM1_TI_Name_wo_comma_and_lines')
        multi_line_any_TM1_TI_name_without_comma.addParseAction(lambda t: ''.join(t))
        
        #26.11.2012, ykud: 
        #Non quoted variables can't have arithmetic signs in them, leaving poor old - out of the above mentioned bunch
        multi_line_any_TM1_TI_unquoted_name_without_comma = Forward()
        any_TM1_TI_unquoted_name_without_comma =  Word( extended_chars + nums + '#}. _$[]').setName('TM1_TI_Name')
        split_any_TM1_TI_unquoted_name_without_comma = any_TM1_TI_unquoted_name_without_comma + Suppress(lineEnd)
        multi_line_any_TM1_TI_unquoted_name_without_comma << ((split_any_TM1_TI_unquoted_name_without_comma + multi_line_any_TM1_TI_unquoted_name_without_comma) | any_TM1_TI_unquoted_name_without_comma).setName('TM1_TI_unquoted_Name_wo_comma_and_lines')
        multi_line_any_TM1_TI_unquoted_name_without_comma.addParseAction(lambda t: ''.join(t))
        
        TI_arithmetic_expression = Forward()
        TI_logical_expression = Forward()
        TI_all_Rules_functions = Forward()
        TI_logical_term = (LPAREN + TI_logical_expression + RPAREN)|(TI_arithmetic_expression + logical_operator + TI_arithmetic_expression) 
        TI_logical_expression << (Optional('~') + TI_logical_term  \
                                + Optional(oneOf('& %') + Group(TI_logical_expression)))
                                                     
        TI_if_function= (Suppress(caseless_literal_with_line_ends("IF")) +  LPAREN + TI_logical_expression + COMMA  + TI_arithmetic_expression + COMMA\
                        + TI_arithmetic_expression +   RPAREN)
        TI_all_Rules_functions << (caseless_literal_with_line_ends(TM1_RULE_FILE_FUNCTIONS.replace('\n',' ').replace('  ',''))
                             + LPAREN + Optional(delimitedList(TI_arithmetic_expression)) + RPAREN ).setParseAction(add_Rules_function)
        #functions that act like keywords (without parameters)
        TI_func_keywords_list = (caseless_literal_with_line_ends(TM1_TI_KEYWORDS) + Optional (LPAREN + RPAREN))

        TI_all_functions = ' '.join(map(lambda t: t.replace('\n',' ').replace('  ',''),
                                    [TM1_TI_ASCII_FUNCTIONS, TM1_TI_ATTRIBUTE_MANIPULATION_FUNCTIONS, TM1_TI_CHORE_FUNCTIONS, TM1_TI_DIMENSION_FUNCTIONS,
                                     TM1_TI_CUBE_FUNCTIONS, TM1_TI_ODBC_FUNCTIONS, TM1_TI_PROCESS_FUNCTIONS, TM1_TI_RULES_FUNCTIONS,
                                     TM1_TI_SANDBOX_FUNCTIONS, TM1_TI_SECURITY, TM1_TI_SERVER_FUNCTIONS, TM1_TI_SUBSET_FUNCTIONS, 
                                     TM1_TI_VIEW_FUNCTIONS, TM1_TI_MISC_FUNCTION]
                                    )
                                    )
        TI_any_function = (
            caseless_literal_with_line_ends(TI_all_functions)
            + LPAREN + Optional(delimitedList(Group(TI_arithmetic_expression))) + RPAREN).setParseAction(add_TI_function)
        
        TI_arithm_term = (Group(TI_any_function)|
                          (LPAREN + TI_arithmetic_expression + RPAREN) |
                       Group(TI_if_function)|Group(TI_all_Rules_functions)|
                        ('!' + multi_line_any_TM1_TI_unquoted_name_without_comma)|
                        ('!' + single_quoted_string)| 
                        caseless_literal_with_line_ends("STET")|caseless_literal_with_line_ends("CONTINUE")|
                        area|
                        single_quoted_string | numeric_constant | multi_line_any_TM1_TI_unquoted_name_without_comma
                      )
        # Not using pypasing.operatorPrecedence since it's uninmportant for drawing links and simple works faster
        TI_arithmetic_expression<< (Optional('-') +  TI_arithm_term 
                                + Optional(Group(oneOf('* \\ / * ^ | + -') + Group(TI_arithmetic_expression))))
        
        #self.test = TI_arithmetic_expression
        # no variables or strings here, since it will eat up all else and elseif
        process_formula = ((LPAREN + TI_arithmetic_expression + RPAREN) | Group(TI_any_function)|Group(TI_if_function)|Group(all_Rules_functions)
                           |TI_func_keywords_list)
        # variable assigment: Variable = formula
        TI_assigment = (any_TM1_TI_name_without_comma + EQUALS + TI_arithmetic_expression).setParseAction(add_TI_variable_to_list)
        
        process_if_expression = Forward()
        process_while_expression = Forward()
        # there can be single statement like '1;', but not 'var;'
        process_statement = process_while_expression|process_if_expression|(process_formula + SEMICOLON)|(TI_assigment + SEMICOLON)|(Word(nums) + SEMICOLON)
        # if statement flow
        process_if_expression << CaselessLiteral('if') + LPAREN + TI_logical_expression + RPAREN + SEMICOLON\
                              + ZeroOrMore (process_statement)\
                              + ZeroOrMore(CaselessLiteral('elseif') + LPAREN + TI_logical_expression + RPAREN + SEMICOLON +  ZeroOrMore (process_statement))\
                              + Optional(CaselessLiteral('else') + SEMICOLON +  ZeroOrMore (process_statement))\
                              + Optional(CaselessLiteral('endif') + SEMICOLON)
        # while statement flow
        process_while_expression << CaselessLiteral('while') +  LPAREN + TI_logical_expression + RPAREN + SEMICOLON\
                                 + OneOrMore(process_statement) + CaselessLiteral('end') + SEMICOLON
        self.test = CaselessLiteral('if') + LPAREN + TI_logical_expression + RPAREN + SEMICOLON
        # File structure 
        # source type and name
        source_type_line = Suppress('562') + COMMA + (DQUOTE +Word(alphas).setResultsName('source_type') + DQUOTE)
        source_name_line = Suppress('586') + COMMA + Optional( QuotedString('"',None,None,False,True).setResultsName('source_name'))
        # process parameters list after 560
        parameter_marker = '560,' + Optional(Word(nums))
        TI_param = (any_TM1_name_without_comma + Suppress(lineEnd)).setParseAction(add_TI_param_to_list)
        #Prolog statement start after 572
        prolog_marker = '572,' + Optional(Word(nums))
        #Metadata statements start after 573
        metadata_marker = '573,' + Optional(Word(nums))
        #Data statements start after 574
        data_marker = '574,' + Optional(Word(nums))
        #Epilog statements start after 575
        epilog_marker = '575,' + Optional(Word(nums))
        process_file_syntax = Suppress(SkipTo(source_type_line)) + source_type_line \
                              + source_name_line \
                              + Suppress(SkipTo(parameter_marker)) + Suppress(parameter_marker) + ZeroOrMore(TI_param)\
                              + '561,' + Optional(Word(nums))\
                              + Suppress(SkipTo(prolog_marker) + prolog_marker) + ZeroOrMore(process_statement) \
                              + Suppress(metadata_marker) + ZeroOrMore(process_statement) \
                              + Suppress(data_marker) + ZeroOrMore(process_statement) \
                              + Suppress(epilog_marker) + ZeroOrMore(process_statement) \
                              + Suppress('576' + SkipTo(stringEnd))
        process_file_syntax.ignore(comment)
        self.process_file_syntax = process_file_syntax
#
def dirList(dir_name, subdir, modified_time = None, *args):
    '''Return a list of file names found in directory 'dir_name'
    if 'subdir' is True, recursively access subdirectories under 'dir_name'.
    if modified_time is set, return only the files newer than given modification time
    Additional arguments, if any, are file extensions to match filenames. Matched
        file names are added to the list.
    If there are no additional arguments, all files found in the directory are
        added to the list.
    Example usage: fileList = dir_list(r'H:\TEMP', False, 'txt', 'py')
        Only files with 'txt' and 'py' extensions will be added to the list.
    Example usage: fileList = dir_list(r'H:\TEMP', True)
        All files and all the files in subdirectories under H:\TEMP will be added
        to the list.
    '''
    fileList = []
    for file in os.listdir(dir_name):
        dirfile = os.path.join(dir_name, file)
        if os.path.isfile(dirfile):
            if len(args) == 0:
                if modified_time and modified_time >= os.path.getmtime(dirfile):
                    pass
                else:
                    fileList.append(dirfile)
            else:
                if os.path.splitext(dirfile)[1][1:].lower() in map(string.lower,args):
                    if modified_time and modified_time >= os.path.getmtime(dirfile):
                        pass
                    else:
                        fileList.append(dirfile)
        # recursively access file names in subdirectories
        elif os.path.isdir(dirfile):
        #    print "Accessing directory:", dirfile
            fileList += dirList(dirfile, subdir, modified_time, *args)
    return fileList
#

def writeToLog(message):
    print ('%s: %s')%(str(datetime.datetime.now()), message)
#
def SafeFileOpen(filename,mode,encoding):
    try:
        fl=codecs.open(filename,mode,encoding,'ignore')
    except IOError:
        print "Cannot open file "+filename+"\n"
        return (-1)
    return fl
    
#
def add_node_to_graph(graph,node_name, node_id):
    ''' Add node to graph, checking whether it exists. Adds required attributes for nodes type drawing'''
    if not graph.has_node(node_id):
        node_attributes = {"id":node_id,"label":('"%s"'%(node_name.rstrip(' '))).encode('utf-8')}
        if node_id[1:4] == 'cub':           
            node_attributes.update({"shape":'box3d',"node_type":"cub"})
        elif node_id[1:4] == 'tm1':
            node_attributes.update({"shape":'box',"node_type":"tm1"})    
        elif node_id[1:4] == 'dim':
            node_attributes.update({"shape":'box',"node_type":"dim"})
        elif node_id[1:4] == 'txt':
            node_attributes.update({"shape":'folder',"style":'dashed',"node_type":"txt"})
        elif node_id[1:5] == 'odbc':
            node_attributes.update({"shape":'component',"style":'dashed',"node_type":"odbc"})
        elif node_id[1:4] == 'pro':
            node_attributes.update({"shape":'diamond',"style":'solid',"color":TI_PRO_COLOR,"fontcolor":TI_PRO_COLOR,"node_type":"pro"})        
        else:
            writeToLog('Unknowh type for node creation')
        graph.add_node(node_id, node_attributes)
#
def add_edge(graph,source,target,edge_id,attr_dict=None):
    '''Add\or search for an edge with given ID'''
    if not (graph.has_edge(source,target,key = edge_id)):
        if attr_dict:
            graph.add_edge(source,target,key = edge_id,**attr_dict)
        else: graph.add_edge(source,target,key = edge_id)
#    
def add_rule_file_content_to_graph(ruleFileName,graph, rule_file_syntax):
    '''
        Add contents of rule file to overall graph.
    '''
    global reference_list
    global feeders_part_of_rule
    global current_node_name
    global parsing_pro_file
    fl = SafeFileOpen(ruleFileName,'r','UTF-8-SIG')
    current_node_name = os.path.basename(ruleFileName).replace(".RUX","").replace(".rux","")
    current_node_id = unique_node_name(current_node_name,'cub')
    add_node_to_graph(graph,current_node_name,current_node_id)
    contents = fl.read()
    # parse file
    reference_list = []
    feeders_part_of_rule = False
    parsing_pro_file = False
    r = rule_file_syntax.parseString(contents,parseAll = True)
    
    for ref in reference_list:
        add_node_to_graph(graph,ref['source_node_name'],ref['source_node_id'])
        add_node_to_graph(graph,ref['target_node_name'],ref['target_node_id'])
        add_edge(graph,ref['source_node_id'],ref['target_node_id'],ref["id"],dict(label = '',style='solid',color='black',edge_type = ''))
        current_label = graph[ref['source_node_id']][ref['target_node_id']][ref['id']]['label']
        if current_label:
            if ref["type"] == 'DB':
                # if we haven't seen this line number already
                if current_label.find(str(ref["line_number"])) < 0:
                    if len(current_label.split('\\n')[-1]) > 80:
                        graph[ref['source_node_id']][ref['target_node_id']][ref['id']]['label'] = current_label +",\\n" + str(ref["line_number"])
                    else:
                        graph[ref['source_node_id']][ref['target_node_id']][ref['id']]['label'] = current_label +"," + str(ref["line_number"])
                else: pass
            if ref["type"] == 'ATTR':
                # if we've already seen this attribute, just add line number to the line
                if [s for s in current_label.split('\\n') if ref["dimension_attribute"]>0]:
                    curr_lbl_list = current_label.split('\\n')
                    lbl_to_add = [s for s in current_label.split('\\n') if ref["dimension_attribute"]>0][0]
                    curr_lbl_list.remove(lbl_to_add)
                    if lbl_to_add.find(str(ref["line_number"])) < 0:
                        lbl_to_add += ', %s' % (ref["line_number"])
                    curr_lbl_list += [lbl_to_add]
                    graph[ref['source_node_id']][ref['target_node_id']][ref['id']]['label'] = (lambda s: '\\n'.join(s))(curr_lbl_list)
                else:
                    graph[ref['source_node_id']][ref['target_node_id']][ref['id']]['label']  = (current_label +
                                                                                                '\\n Attribute %s from stmt line %s'% (ref["dimension_attribute"],str(ref["line_number"])))
        else:
            if ref["type"] == 'DB':
                if ref["feeder"]:
                    graph[ref['source_node_id']][ref['target_node_id']][ref['id']]['label']  = 'Feeder line %s' % (str(ref["line_number"]))
                    graph[ref['source_node_id']][ref['target_node_id']][ref['id']]['edge_type'] = 'Feeder'
                else:
                    graph[ref['source_node_id']][ref['target_node_id']][ref['id']]['label']  = 'DB line %s' % (str(ref["line_number"]))
                    graph[ref['source_node_id']][ref['target_node_id']][ref['id']]['edge_type'] = 'DB'
            if ref["type"] == 'ATTR':
                graph[ref['source_node_id']][ref['target_node_id']][ref['id']]['label']  = 'Attribute %s from stmt line %s'% (ref["dimension_attribute"].encode('utf-8')
                                                                                                                              ,str(ref["line_number"]))
                graph[ref['source_node_id']][ref['target_node_id']][ref['id']]['edge_type'] = 'Attr'
                
        graph[ref['source_node_id']][ref['target_node_id']][ref['id']]['color']  = ref["color"]
        graph[ref['source_node_id']][ref['target_node_id']][ref['id']]['style']  = ref["arrow_style"]
#
def search_dict (variable_dict, key):
    '''Trying to find lowest level assigment for given dict key. Returns value or false if key not found'''
    if key in variable_dict:
        if not search_dict(variable_dict,variable_dict[key]):
            return variable_dict[key]
        else:
            return search_dict(variable_dict,variable_dict[key])
    else:
        return False
#
def add_process_file_content_to_graph(process_file_name,graph, process_file_syntax):
    '''
        Add contents of process  file to overall graph.
    '''
    global reference_list
    global TI_variable_list
    global current_node_name
    global parsing_pro_file
    fl = SafeFileOpen(process_file_name,'r','UTF-8-SIG')
    process_name = os.path.basename(process_file_name).replace(".pro","")
    contents = fl.read()
    # parse file
    reference_list = []
    TI_variable_list = {}
    parsing_pro_file = True
    current_node_name = process_name
    current_node_id = unique_node_name(process_name,'pro')
    add_node_to_graph(graph,process_name,current_node_id)
    r = process_file_syntax.parseString(contents,parseAll = True)
    source_node = None
    # checking explicit process source
    #print r["source_type"]
    if r["source_type"]=='VIEW':            
       source_node_id = unique_node_name(r["source_name"],'cub')
       add_node_to_graph(graph,r["source_name"],source_node_id)
    elif r["source_type"] == 'CHARACTERDELIMITED':
        source_node_id = unique_node_name(r["source_name"],'txt')
        # if there is a / in file name, need to tripple it to appear in drawing )
        add_node_to_graph(graph,r["source_name"].replace('\\','\\\\\\'),source_node_id)
    elif r["source_type"] == 'ODBC':
        source_node_id = unique_node_name(r["source_name"],'odbc')
        add_node_to_graph(graph,r["source_name"],source_node_id)
    elif r["source_type"] == 'SUBSET':
        source_node_id = unique_node_name(r["source_name"],'dim')
        add_node_to_graph(graph,r["source_name"],source_node_id)
    elif r["source_type"] == 'NULL':
        source_node_id = unique_node_name(process_name,'pro')
        add_node_to_graph(graph,process_name,source_node_id) 
    elif r["source_type"] == 'ODBODIMENSION':
        source_node_id = unique_node_name(r["source_name"],'tm1')
        add_node_to_graph(graph,r["source_name"],source_node_id)    
    else:
        pass
    # drawing Source to Pro edge
    #print source_node_id, process_name
    if source_node_id <> unique_node_name(process_name,'pro'):
        edg_id ='"%s--->%s"'%(source_node_id.replace('"',''),current_node_id.replace('"',''))
        add_edge(graph, source_node_id,current_node_id, edg_id, dict(style='solid', color=TI_PRO_COLOR, fontcolor = TI_PRO_COLOR,
                         label=('"Source %s for process %s. "'%(r["source_name"],process_name)).encode('utf-8'),edge_type='pro'))
    for ref in reference_list:  
        # Calculating Target Node
        # checking whether target node is a variable and we can substitute it with real value
        if search_dict(TI_variable_list,ref['target_node_name']):
        # If it's a simple value -- substitute, otherwise, link it back to process box
            if search_dict(TI_variable_list,ref['target_node_name']) == 'Calculation':
                target_node_id = current_node_id
                ref['type'] = 'Calculated ' + ref['type']
            #checking whether node is a parameter passed to this process
            elif search_dict(TI_variable_list,ref['target_node_name']) == 'PARAMETER':
                target_node_id = current_node_id
                ref['type'] = 'Parameter selected for ' + ref['type']
            else:
                target_node_id = unique_node_name(search_dict(TI_variable_list,ref['target_node_name']),ref['target_node_type'])
                add_node_to_graph(graph,search_dict(TI_variable_list,ref['target_node_name']).replace('\\','\\\\\\'),target_node_id)   
        else:
            target_node_id = unique_node_name(ref['target_node_name'],ref['target_node_type'])
            add_node_to_graph(graph,ref['target_node_name'].replace('\\','\\\\\\'),target_node_id)
        # Calculating Target Node
        # checking whether source node is a variable and we can substitute it with real value
        if search_dict(TI_variable_list,ref['source_node_name']):
        # If it's a simple value -- substitute, otherwise, link it back to process box
            if search_dict(TI_variable_list,ref['source_node_name']) == 'Calculation':
                source_node_id = current_node_id
                ref['type'] = 'Calculated ' + ref['type']
            #checking whether node is a parameter passed to this process
            elif search_dict(TI_variable_list,ref['source_node_name']) == 'PARAMETER':
                source_node_id = current_node_id
                ref['type'] = 'Parameter selected for ' + ref['type']
            else:
                source_node_id = unique_node_name(search_dict(TI_variable_list,ref['source_node_name']),ref['source_node_type'])
                add_node_to_graph(graph,search_dict(TI_variable_list,ref['source_node_name']), source_node_id)   
        else:

            source_node_id = unique_node_name(ref['source_node_name'],ref['source_node_type'])
            add_node_to_graph(graph,ref['source_node_name'],source_node_id)
        
        # adding edge
        # checking if edge already exists
        edge_id = '"pro_%s--->%s"'%(source_node_id.replace('"',''),target_node_id.replace('"',''))
        add_edge(graph, source_node_id, target_node_id, edge_id
                     ,dict(style=ref["arrow_style"],color=ref["color"], fontcolor = ref["color"], label='',edge_type='pro'))
        # setting up label
        current_label = graph[source_node_id][target_node_id][edge_id]["label"].replace('"','')
        if current_label.find(ref['type'].encode('utf-8')) < 0:
            if len(current_label.split('\\n')[-1]) > 80:
                current_label += '\\n%s ' % (ref['type'].encode('utf-8'))
            else:
                current_label += '%s ' % (ref['type'].encode('utf-8'))
            graph[source_node_id][target_node_id][edge_id]["label"] = '"%s"'%(current_label)
    #'''
#  
def parse_rules_file(contents):
    '''parse a given rules file contents. used for testing'''
    syntax = TM1Syntax()
    rs = syntax.rule_file_syntax
    global reference_list
    global feeders_part_of_rule
    global current_node_name
    global parsing_pro_file
    parsing_pro_file = False
    reference_list = []
    feeders_part_of_rule = False
    current_node_name = 'test'
    r = rs.parseString(contents,parseAll = True)
#
def parse_process_file(contents):
    '''parse a given process file contents. used for testing'''
    syntax = TM1Syntax()
    ps = syntax.process_file_syntax
    global reference_list
    global TI_variable_list
    global parsing_pro_file
    global current_node_name
    current_node_name = 'test'
    parsing_pro_file = True
    TI_variable_list = {}
    reference_list = []
    p = ps.parseString(contents,parseAll = True)
    print ('%s')%(p.dump())
    print ('SourceType: %s, SourceName: %s')%(p.source_type, p.source_name)
    print ('Reference_List: %s') % (reference_list)
    print ('Variable_List: %s') % (TI_variable_list)
#
def generate_graph_metadata(InDirectory, forceRebuild = False):
    """ Parses a model in InDirectory and creates networkx graph out of it.
        If forceRebuild flag is set, model is reprocessed fully
    """
    graph_file_name = InDirectory.replace('\\','').replace(':','').replace('/','').replace('.','')
    starttime_full=time.time()
    # Should get rid of incremental ideas all together, make it more robust
    graph = nx.MultiDiGraph(graph = {"directed":True,"simplify":False,"rankdir":'LR',
                                     "name":graph_file_name, "scan_time":0, "scan_datetime":0})
    graph.graph['graph']["scan_time"] = 1
    
    # check if such graph already exists
    #if os.path.isfile('%s.gz'%(graph_file_name)):
        #print graph_file_name
    #    graph = nx.readwrite.gpickle.read_gpickle("%s.gz"%(graph_file_name))
    #else:
    #    graph = nx.MultiDiGraph(graph = {"directed":True,"simplify":False,"rankdir":'LR',
    #                                 "name":graph_file_name, "scan_time":0, "scan_datetime":0})
    #if forceRebuild:
    #     graph.graph['graph']["scan_time"] = 1
         
    num_of_files = 0
    num_of_errors = 0
    file_list = dirList( unicode(InDirectory) , 1,  graph.graph['graph']["scan_time"] , 'cub' , 'RUX', 'pro')
    if file_list:
        # Need to add incremental processing here, it should
        # remove non-existent cubes
        # remove non-existent processes
        # for changed processes: remove changed process nodes and reprocess pro files
        # for changed rules: delete all links in cube, except Feeders
        #                    delete feeder links from cube
        

        # adding all cubes as graph nodes first
        for f in dirList( unicode(InDirectory) , 1, 1 ,'cub'):
            # ignoring system cubes
            if f.find('}') < 1:
                add_node_to_graph(graph,unicode(os.path.basename(f).replace(".cub","")), unique_node_name(os.path.basename(f).replace(".cub","")))
        # drawing rules
        #'''
        syntax = TM1Syntax()
        rs = syntax.rule_file_syntax
        for f in dirList( unicode(InDirectory) , 1, 1 ,'RUX'):
            # ignoring system rules
            if f.find('}') < 1:
                writeToLog('Parsing file %s' % (f) )
                try:
                    add_rule_file_content_to_graph(f,graph,rs)
                except:
                    writeToLog('Failed to parse file: %s, please send it to developers for fixing this bug'%(f))
                    num_of_errors += 1
                
            num_of_files += 1
        # drawing processes
        ps = syntax.process_file_syntax
        for f in dirList( unicode(InDirectory) , 1, 1 ,'pro'):
            # ignoring system processes
            if f.find('}') < 1:
                writeToLog('Parsing file %s' % (f) )
                try:
                    add_process_file_content_to_graph(f,graph,ps)
                except:
	                writeToLog('Failed to parse file: %s, please send it to developers for fixing this bug'%(f))
	                #writeToLog('Exception stack trace'%(sys.exc_info()[0]))
	                print sys.exc_info()
	                num_of_errors += 1
	            #continue:
            num_of_files += 1
        
        writeToLog("Resulting graph contains %d nodes and %d edges"%(graph.number_of_nodes(),graph.size()))
        writeToLog("Pro Edges %d"%(len ([ (u,v,d) for u,v,d in graph.edges(data=True) if d ['edge_type']=='pro'])))
        writeToLog("DB Edges %d"%(len ([ (u,v,d) for u,v,d in graph.edges(data=True) if d ['edge_type']=='DB'])))
        writeToLog("Feeder Edges %d"%(len ([ (u,v,d) for u,v,d in graph.edges(data=True) if d ['edge_type']=='Feeder'])))
        writeToLog("Attr Edges %d"%(len ([ (u,v,d) for u,v,d in graph.edges(data=True) if d ['edge_type']=='Attr'])))
        if num_of_errors:
            writeToLog ('%s ERROR(S) FOUND DURING PARSING. Please scroll up the log and contact developers to fix this bug. It''ll help all of us ) Thank you.'%(num_of_errors))
        graph.graph['graph']["scan_time"] = starttime_full
        graph.graph['graph']["scan_datetime"] = datetime.datetime.now()
        nx.readwrite.gpickle.write_gpickle(graph, "%s.gz"%(graph_file_name))
        writeToLog("Processed %d files in " %(num_of_files) + '%.2f'%(time.time()-starttime_full) + " seconds.")    
    else:
        writeToLog("No new or updated files found, graph is up to date")
    return graph
    
#

def export_graph_to_csv (graph):
    '''Exporting graph contents to csv file for basic Excel analysis'''
    out_put_fl = SafeFileOpen('graph_contents.csv',"wb",'utf-8') #-sig
    out_put_fl.write('SourceType;Source;TargetType;Target;LinkType;Link Details\n')
    for u,v,d in graph.edges(data=True):
         out_put_fl.write('%s;%s;%s;%s;%s;%s\n'%
                          (graph.node[u]['node_type'].decode('utf-8') ,graph.node[u]['label'].decode('utf-8'),
                           graph.node[v]['node_type'].decode('utf-8') ,graph.node[v]['label'].decode('utf-8'),
                           d['edge_type'], d['label'].decode('utf-8')
                           ))    
    out_put_fl.close()
def main (argv):
    '''input format: python tm1mn.py -i c:\input\ -l log_file.txt
    Parses TM1 model in given directory, creates a networkx graph and writes model contents to csv file
    Arguments:
    -i or --input TM1 server directory, containing rux cub and pro files
    -l or --log log file (if ommited, log will be displayed in console)
    -h or --help this message
    '''
    WorkLogFile=""#"tm1mn.log"
    InDirectory = ""
    outputFile = ""
    cubeName = ""
    # getting command line arguments
    try:
        opts, args = getopt.getopt(argv, "hi:l:", ["help","input=","","log="])
    except getopt.GetoptError:
        print main.__doc__
        sys.exit(2)
    if len(argv) <= 1:
        print main.__doc__
        sys.exit(2)
    for opt,arg in opts:
        if opt in ("-h","--help"):
            print main.__doc__
            sys.exit(2)
        elif opt in ("-i","--input"):
            InDirectory = arg
        elif opt in ("-l","--log"):
            WorkLogFile = arg
 
    if WorkLogFile!="":
        WorkLogFilefl = SafeFileOpen(WorkLogFile,"w",'utf-8')
        backup_stdout = sys.stdout
        backup_stderr = sys.stderr
        sys.stdout = WorkLogFilefl
        sys.stderr = WorkLogFilefl

    writeToLog("Starting scanning.")
    writeToLog("Input Directory is: %s. Log file is: %s"%(InDirectory, WorkLogFile))

    gr = generate_graph_metadata(InDirectory)
    export_graph_to_csv(gr)
    nx.drawing.nx_pydot.write_dot(gr,"graph_to_draw.dot")


    if WorkLogFile!="":
        WorkLogFilefl.close()
        sys.stdout = backup_stdout
        sys.stderr = backup_stderr

if __name__ == "__main__":
    main(sys.argv[1:])
